from fpocketr.rnavigate.plots.functions.functions import get_contrasting_colors

from fpocketr.rnavigate.plots.functions.ss import (
    plot_annotation_ss,
    plot_basepairs_ss,
    plot_interactions_ss,
    plot_nucleotides_ss,
    plot_positions_ss,
    plot_sequence_ss,
    plot_structure_ss,
)

__all__ = [
    "get_contrasting_colors",
    "plot_annotation_ss",
    "plot_basepairs_ss",
    "plot_interactions_ss",
    "plot_nucleotides_ss",
    "plot_positions_ss",
    "plot_sequence_ss",
    "plot_structure_ss",
]


"""Classes for storing and manipulating data for RNAvigate.

This module contains the base classes for RNAvigate data classes:
    Sequence: represents a nucleotide sequence
    Data: represents a data table with a sequence
"""

from os.path import isfile
import Bio.SeqIO
import numpy as np
import pandas as pd
import matplotlib.colors as mpc
from fpocketr.rnavigate import styles
from fpocketr.rnavigate import data


def normalize_sequence(sequence, t_or_u="U", uppercase=True):
    """Returns sequence as all uppercase nucleotides and/or corrects T or U.

    Parameters
    ----------
    sequence : string or RNAvigate Sequence)
        The sequence
        If given an RNAvigate Sequence, the sequence string is retrieved
    t_or_u : "T", "U", or False, defaults to "U"
        "T" converts "U"s to "T"s
        "U" converts "T"s to "U"s
        False does nothing
    uppercase bool, defaults to True
        Whether to make sequence all uppercase

    Returns
    -------
        string
            the cleaned-up sequence string
    """
    if isinstance(sequence, Sequence):
        sequence = sequence.sequence
    if uppercase:
        sequence = sequence.upper()
    if not t_or_u:
        pass
    elif t_or_u.upper() == "U":
        sequence = sequence.replace("t", "u").replace("T", "U")
    elif t_or_u.upper() == "T":
        sequence = sequence.replace("u", "t").replace("U", "T")
    return sequence


class Sequence:
    """A class for storing and manipulating RNA sequences.

    Parameters
    ----------
    sequence : string or pandas.DataFrame
        sequence string, fasta file, or a Pandas dataframe containing a
        "Sequence" column
    name : string, optional
        The name of the sequence, defaults to None

    Attributes
    ----------
    sequence : string
        The sequence string
    name : string
        The name of the sequence
    other_info : dict
        A dictionary of additional information about the sequence
    null_alignment : SequenceAlignment
        An alignment of the sequence to itself
    """

    def __init__(self, input_data, name=None):
        """Initialize the Sequence object."""
        self.name = name
        self.other_info = dict()
        if isinstance(input_data, str):
            if isfile(input_data):
                self.sequence = self.read_fasta(input_data)
            else:
                self.sequence = input_data
        elif isinstance(input_data, pd.DataFrame):
            self.sequence = self.get_seq_from_dataframe(input_data)
        elif isinstance(input_data, Sequence):
            self.sequence = input_data.sequence
        if "-" not in self.sequence:
            self.null_alignment = data.SequenceAlignment(self, self)

    @classmethod
    def from_fasta(cls, fasta, entry=0):
        """Create a Sequence object from a fasta file.

        Parameters
        ----------
        fasta : string
            path to fasta file
        entry : int, defaults to 0
            the index of the sequence in the fasta file

        Returns
        -------
        sequence : rnavigate.data.Sequence
            the sequence object
        """
        with open(fasta, "r") as file:
            fasta = list(Bio.SeqIO.parse(file, "fasta"))
        return cls(str(fasta[entry].seq).replace("T", "U"))

    def __str__(self):
        """Return the name of the sequence."""
        if self.name is None:
            return "seq-object"
        return self.name

    def read_fasta(self, fasta):
        """Parse a fasta file for the first sequence.

        Parameters
        ----------
        fasta : string
            path to fasta file
        """
        with open(fasta, "r") as file:
            fasta = list(Bio.SeqIO.parse(file, "fasta"))
        return str(fasta[0].seq).replace("T", "U")

    def get_seq_from_dataframe(self, dataframe):
        """Parse a dataframe for the sequence string, store as self.sequence.

        Parameters
        ----------
        dataframe : pandas.DataFrame
            must contain a "Sequence" column
        """
        sequence = "".join(dataframe["Sequence"].values)
        return sequence.replace("T", "U").replace("t", "u")

    @property
    def length(self):
        """Get the length of the sequence

        Returns
        -------
        length : int
            the length of self.sequence
        """
        return len(self.sequence)

    def get_region(self, region="all"):
        """Checks region input for validity and returns start and end positions.

        If region is "all", returns 1, self.length. Otherwise, ensures that region
        is between these values and returns the values, sorted.

        Parameters
        ----------
        region : list of 2 int
            start and end positions of the region

        Returns
        -------
        start, end : int, int
            the starting and ending positions
        """
        if region == "all":
            return 1, self.length
        start, end = region
        if not (1 <= start <= self.length) or not (1 <= end <= self.length):
            raise ValueError(
                f"Region {region} is outside of sequence range (1, {self.length})"
            )
        return sorted([start, end])

    def normalize_sequence(self, t_or_u="U", uppercase=True):
        """Converts sequence to all uppercase nucleotides and corrects T or U.

        Parameters
        ----------
        t_or_u : "T", "U", or False, defaults to "U"
            "T" converts "U"s to "T"s
            "U" converts "T"s to "U"s
            False does nothing.
        uppercase : bool, defaults to True
            Whether to make sequence all uppercase
        """
        self.sequence = normalize_sequence(self, t_or_u=t_or_u, uppercase=uppercase)

    def get_region_data(self, region="all"):
        """Get a copy of the data object containing only the specified region.

        Parameters
        ----------
        region : list of 2 int, defaults to "all"
            start and end positions of the region

        Returns
        -------
        region_data : rnavigate.data.Sequence
            the sequence containing only the specified region
        """
        start, end = self.get_region(region)
        target_sequence = self.sequence[start - 1 : end]
        data.set_alignment(
            sequence1=self.sequence,
            sequence2=target_sequence,
            alignment1=self.sequence,
            alignment2="-" * (start - 1) + target_sequence + "-" * (self.length - end),
        )
        alignment = data.SequenceAlignment(self.sequence, target_sequence)
        return self.get_aligned_data(alignment)

    def get_aligned_data(self, alignment):
        """Get a copy of the sequence positionally aligned to another sequence.

        Parameters
        ----------
        alignment : rnavigate.data.Alignment
            the alignment to use

        Returns
        -------
        aligned_sequence : rnavigate.data.Sequence
            the aligned sequence
        """
        return Sequence(alignment.target_sequence)

    def get_colors_from_sequence(self):
        """Get a colors and colormap representing the nucleotide sequence.

        Returns
        -------
        colors : numpy array
            one matplotlib color-like value for each nucleotide in self.sequence
        colormap : rnavigate.data.ScalarMappable
            a colormap used for creating a colorbar
        """
        colors = np.array([styles.get_nt_color(nt) for nt in self.sequence])
        colormap = data.ScalarMappable(
            cmap=[styles.get_nt_color(nt) for nt in "AUGC"],
            normalization="none",
            values=None,
            extend="neither",
            title="Nucleotide identity",
            alpha=1,
            ticks=[0, 1, 2, 3],
            tick_labels=["A", "U", "G", "C"],
        )
        return colors, colormap

    def get_colors_from_positions(self, pos_cmap="rainbow"):
        """Get colors and colormap representing the nucleotide position.

        Parameters
        ----------
        pos_cmap : str, defaults to "rainbow"
            cmap used for position colors

        Returns
        -------
        colors : numpy array
            one matplotlib color-like value for each nucleotide in self.sequence
        colormap : rnavigate.data.ScalarMappable
            a colormap used for creating a colorbar
        """
        colormap = data.ScalarMappable(
            cmap=pos_cmap,
            normalization="min_max",
            values=[1, self.length],
            extend="neither",
            title="Nucleotide position",
            alpha=1,
        )
        colors = colormap.values_to_hexcolors(np.arange(self.length))
        return colors, colormap

    def get_colors_from_profile(self, profile):
        """Get colors and colormap representing per-nucleotide data.

        Parameters
        ----------
        profile : rnavigate.data.Profile
            the profile to use to get colors.

        Returns
        -------
        colors : numpy array
            one matplotlib color-like value for each nucleotide in self.sequence
        colormap : rnavigate.data.ScalarMappable
            a colormap used for creating a colorbar
        """
        alignment = data.SequenceAlignment(profile, self)
        colors = alignment.map_values(profile.colors, fill="#808080")
        colormap = profile.cmap
        return colors, colormap

    def get_colors_from_annotations(self, annotations, default_color="gray"):
        """Get colors and colormap representing sequence annotations.

        Parameters
        ----------
        annotations : list of rnavigate.data.Annotations
            the annotations to use to get colors.
        default_color : matplotlib color-like, defaults to "gray"
            the color to use for nucleotides not in any annotation

        Returns
        -------
        colors : numpy array
            one matplotlib color-like value for each nucleotide in self.sequence
        colormap : rnavigate.data.ScalarMappable
            a colormap used for creating a colorbar
        """
        colors = np.full(self.length, default_color, dtype="<U16")
        cmap = [default_color]
        tick_labels = ["other"]
        for annotation in annotations:
            cmap.append(annotation.color)
            tick_labels.append(annotation.name)
            annotation = annotation.get_aligned_data(
                data.SequenceAlignment(annotation, self)
            )
            for site in annotation.get_sites():
                colors[site - 1] = annotation.color
        colormap = data.ScalarMappable(
            cmap=cmap,
            normalization="none",
            values=None,
            extend="neither",
            title="Annotations",
            alpha=1,
            ticks=list(range(len(annotations) + 1)),
            tick_labels=tick_labels,
        )
        return colors, colormap

    def get_colors_from_structure(self, structure):
        """Get colors and colormap representing base-pairing status.

        Parameters
        ----------
        structure : rnavigate.data.SecondaryStructure
            the structure to use to get colors.

        Returns
        -------
        colors : numpy array
            one matplotlib color-like value for each nucleotide in self.sequence
        colormap : rnavigate.data.ScalarMappable
            a colormap used for creating a colorbar
        """
        cmap = ["darkOrange", "darkOrchid", "gray"]
        ct_colors = [cmap[int(pair == 0)] for pair in structure.pair_nts]
        alignment = data.SequenceAlignment(structure, self)
        colors = alignment.map_values(ct_colors, fill="gray")
        colormap = data.ScalarMappable(
            cmap=cmap,
            normalization="none",
            values=None,
            extend="neither",
            title="Base-pairing status",
            alpha=1,
            ticks=[0, 1, 2],
            tick_labels=["unpaired", "paired", "unaligned"],
        )
        return colors, colormap

    def get_colors(
        self, source, pos_cmap="rainbow", profile=None, structure=None, annotations=None
    ):
        """Get colors and colormap representing information about the sequence.

        Parameters
        ----------
        source : str, list, or matplotlib color-like
            the source of the color information
            if a string, must be one of:
                "sequence", "position", "profile", "structure", "annotations"
            if a list, must be a list of matplotlib color-like values, colormap
                will be None.
            if a matplotlib color-like value, all nucleotides will be colored
                that color, colormap will be None.
        pos_cmap : str, defaults to "rainbow"
            cmap used for position colors if source is "position"
        profile : rnavigate.data.Profile, optional
            the profile to use to get colors if source is "profile"
        structure : rnavigate.data.SecondaryStructure, optional
            the structure to use to get colors if source is "structure"
        annotations : list of rnavigate.data.Annotations, optional
            the annotations to use to get colors if source is "annotations"

        Returns
        -------
        colors : numpy array
            one matplotlib color-like value for each nucleotide in self.sequence
        colormap : rnavigate.data.ScalarMappable
            a colormap used for creating a colorbar
        """
        if isinstance(source, str) and (source == "sequence"):
            return self.get_colors_from_sequence()
        elif isinstance(source, str) and (source == "position"):
            return self.get_colors_from_positions(pos_cmap=pos_cmap)
        elif isinstance(source, str) and (source == "profile"):
            return self.get_colors_from_profile(profile=profile)
        elif isinstance(source, str) and (source == "annotations"):
            return self.get_colors_from_annotations(annotations=annotations)
        elif isinstance(source, str) and (source == "structure"):
            return self.get_colors_from_structure(structure)
        elif mpc.is_color_like(source):
            return np.full(self.length, source, dtype="<U16"), None
        elif (len(source) == self.length) and all(mpc.is_color_like(c) for c in source):
            return np.array(list(source)), None
        else:
            print("Invalid colors:")
            print("\tchoices are profile, sequence, position, structure,")
            print("annotations, a list of mpl colors, or a single mpl color.")
            print("Defaulting to sequence.")
            return self.get_colors_from_sequence()


class Data(Sequence):
    """The base class for RNAvigate Profile and Interactions classes.

    Parameters
    ----------
    input_data : pandas.DataFrame or str
        a pandas dataframe or path to a data file
    sequence : string or rnavigate.data.Sequence
        the sequence to use for the data
    metric : string or dict
        the column of the dataframe to use as the default metric to visualize
    metric_defaults : dict
        a dictionary of metric defaults
    read_table_kw : dict, optional
        kwargs dictionary passed to pd.read_table
    name : string, optional
        the name of the data, defaults to None

    Attributes
    ----------
    data : pandas.DataFrame
        the data table
    filepath : string
        the path to the data file
    sequence : string or rnavigate.data.Sequence
        the sequence to use for the data
    metric : string or dict
        the column of the dataframe to use as the metric to visualize
    metric_defaults : dict
        A dictionary of metric values and default settings for visualization
    default_metric : string
        the default metric to use for visualization
    """

    def __init__(
        self,
        input_data,
        sequence,
        metric,
        metric_defaults,
        read_table_kw=None,
        name=None,
    ):
        """Initialize the Data object."""
        if read_table_kw is None:
            read_table_kw = {}
        # assign data
        if isinstance(input_data, pd.DataFrame):
            self.data = input_data
            self.filepath = "dataframe"
        elif isfile(input_data):
            self.data = self.read_file(input_data, read_table_kw)
            self.filepath = input_data
        elif isinstance(input_data, str):
            raise ValueError(f"file path ({input_data}) not found.")
        else:
            raise ValueError("input_data must be pandas DataFrame or valid file path")
        # assign sequence
        if sequence is None:
            sequence = self.data
        super().__init__(sequence, name=name)
        # assign metrics
        self.metric_defaults = {
            "default": {
                "metric_column": "Profile",
                "error_column": None,
                "color_column": None,
                "cmap": "viridis",
                "normalization": "0_1",
                "values": None,
                "extend": "neither",
                "title": "Type: metric",
                "alpha": 0.7,
            },
            "Distance": {
                "metric_column": "Distance",
                "error_column": None,
                "color_column": None,
                "cmap": "cool",
                "normalization": "min_max",
                "values": [5, 50],
                "extend": "both",
                "title": "3D distance",
                "alpha": 0.7,
            },
        }
        self.add_metric_defaults(metric_defaults)
        self.default_metric = metric
        self._metric = None
        self.metric = metric

    def add_metric_defaults(self, metric_defaults):
        """Add metric defaults to self.metric_defaults"""
        default_defaults = self.metric_defaults["default"]
        for metric, defaults in metric_defaults.items():
            self.metric_defaults[metric] = default_defaults | defaults

    @property
    def metric(self):
        """Get the column of the dataframe to use as the metric for visualization."""
        return self._metric["metric_column"]

    @metric.setter
    def metric(self, value):
        """Set the column of the dataframe and other settings to use for visualization.

        Parameters
        ----------
        value : string or dict
            the column of the dataframe to use as the metric for visualization
            if a string, must be one of:
                "default", "Distance", or a column name in self.data
            if a dict, must contain at least one of:
                "metric_column", "error_column", "color_column", "cmap",
                "normalization", "values", "extend", "title", or "alpha"
        """
        if isinstance(value, str):
            try:
                self._metric = self.metric_defaults[value]
                return
            except KeyError as exception:
                if value not in self.data.columns:
                    print(
                        f"metric ({value}) not found in data:\n"
                        + str(list(self.data.columns))
                    )
                    raise exception
                self._metric = self.metric_defaults["default"]
                self._metric["metric_column"] = value
                return
        try:
            defaults = self.metric_defaults[value["metric_column"]]
        except KeyError:
            defaults = self.metric_defaults["default"]
        for key in value:
            if key not in defaults:
                raise ValueError(
                    f"{key} is not an expected value for metric "
                    f"setting:\n{list(defaults.keys())}"
                )
        self._metric = defaults | value

    @property
    def error_column(self):
        """Get the column of the dataframe to use as the error for visualization."""
        if self._metric["error_column"] in self.data.columns:
            return self._metric["error_column"]
        print(f"Warning: {self} missing expected error column")
        return None

    @property
    def color_column(self):
        """Get the column of the dataframe to use as the color for visualization."""
        if self._metric["color_column"] in self.data.columns:
            return self._metric["color_column"]
        return self._metric["metric_column"]

    @property
    def cmap(self):
        """Get the colormap to use for colorbars and to retrieve colors."""
        cmap_kwargs = {}
        for k, v in self._metric.items():
            if not k.endswith("column"):
                cmap_kwargs[k] = v
        return data.ScalarMappable(**cmap_kwargs)

    @property
    def colors(self):
        """Get one matplotlib color-like value for each nucleotide in self.sequence."""
        values = self.data[self.color_column]
        if values.dtype.name == "bool":
            values = values.astype("int")
        return self.cmap.values_to_hexcolors(np.ma.masked_invalid(values))

    def read_file(self, filepath, read_table_kw):
        """Convert data file to pandas dataframe and store as self.data

        Parameters
        ----------
        filepath : string
            path to data file containing interactions
        read_table_kw : dict
            kwargs dictionary passed to pd.read_table

        Returns
        -------
        dataframe : pandas.DataFrame
            the data table
        """
        return pd.read_table(filepath, **read_table_kw)



"""RNAvigate

RNA visualization and graphical analysis toolset

A Jupyter-compatible toolset for visually exploring RNA structure and chemical
probing data.
"""

# from rnavigate.rnavigate import Sample
# from rnavigate.plotting_functions import (
#     plot_options,
#     plot_ss,
# )

# from rnavigate import data
# from rnavigate import plots
# from rnavigate import styles

# __version__ = "1.0.0"
# __author__ = "Patrick S. Irving"
# __email__ = "psirving@email.unc.edu"

# __all__ = [
#     "Sample",
#     # plotting functions
#     "plot_options",
#     "plot_ss",
#     # modules
#     "data",
#     "plots",
#     "styles",
# ]



#!/usr/bin/env python3
# -----------------------------------------------------------------------------
# Module for parsing fpocket outputs and analyzing pocket characteristics
# Seth Veenbaas
# Weeks Lab, UNC-CH
# 2022
#
# Version 1.0.1
#
# -----------------------------------------------------------------------------
import os
import math
from re import findall, sub
from prody import *
import numpy as np
import pandas as pd
import requests
from pymol import cmd
from rdkit import Chem
from rdkit.Chem import QED
import trimesh
from fpocketr import util
from prody.utilities import openFile


def analyze_pockets(
    pdb : str,
    pqr_out : str,
    pdb_out : str,
    analysis : str,
    name : str,
    info_txt : str,
    pockets_out : list[str],
    pdb_code : str,
    chain : str,
    state,
    ligandchain : str,
    ligand : str,
    m : float,
    M : float,
    i : int,
    D : float,
    A : int,
    p : float,
    qualityfilter : float,
    knownnt : list[int],
    ) -> tuple[pd.DataFrame, prody.AtomGroup]:

    # Parses pdb files and returns prody structure objects.
    ligand_rna_structure = parsePDB(pdb)
    out_rna_structure = parsePDB(pdb_out)
    pocket_structure = parsePQR(pqr_out)

    # Sets ligand chain to first pdb chain by default.
    if ligandchain is None:
        ligandchain = chain[0]

    # Create real_sphere.pdb ouput be combinding the pqr_out and pdb_out.
    get_real_sphere(out_rna_structure, pocket_structure, analysis, name, pqr_out)

    # Creates a dataframe with pocket characteristics from fpocket info.txt
    pc_df = get_characteristics(
        info_txt,
        pdb_code,
        name,
        m,
        M,
        i,
        D,
        A,
        p,
        state,
    )

    rna_coords = out_rna_structure.copy()

    # If pockets are detected calculate features and add to pc_df.
    if out_rna_structure.select('resname STP'):

        # Get atomgroup for rna and add pocket characteristics.
        stp_coords = out_rna_structure.select('resname STP').copy()

        add_basic_characteristics(
            stp_coords,
            pockets_out,
            qualityfilter,
            pc_df,
            chain,
            analysis,
            name,
            knownnt,
        )
        
        # Get atomgroup for ligand and add ligand characteristics.
        if ligand in ('n', 'N', 'no', 'No', 'none', 'None'):
            ligand_coords = None
            ligand = None
        else:
            ligand_coords, ligand = get_ligand_coords(
                ligand_rna_structure,
                ligand,
                ligandchain,
                analysis,
                name,
            )
            if ligand_coords:
                add_ligand_characteristics(
                    analysis,
                    stp_coords,
                    ligand_coords,
                    ligand,
                    pc_df,
                    knownnt,
                )

    return pc_df, rna_coords


# -----------------------------------------------------------------------------


def reformat_pqr(filename : str):
    """Reformats PQR files to add whitespace between all data fields.
    This is required for Prody 2.4 to properly parse PQR files.
    NOTE: reformated PQR files will not be read properly by PyMol 2.5.

    Args:
        filename (str): Path to PQR file.

    Returns:
        str: Path to reformated PQR file. <filename>_prody.pqr
    """
    pqr = openFile(filename, 'rt')
    lines = pqr.readlines()
    pqr.close()
    reformat=[]
    atom = False

    for line in lines:
        startswith = line[0:6].strip()

        if startswith == 'ATOM' or startswith == 'HETATM':
            atom = True
            fields = line.split()
            if fields[5].find('.') != -1:
            # coords too early as no chid
                fields.insert(4, '')
            if len(fields) != 11:
                try:
                    fields = fields[:6] + [line[30:38].strip(), line[38:46].strip(), line[46:54].strip()] + line[54:].split()
                except:
                    continue
            
            serial_str = fields[1]
            atomname= fields[2]
            resname = fields[3]
            chid = fields[5]
            alt = ' '
            coordinates_x = fields[6]
            coordinates_y = fields[7]
            coordinates_z = fields[8]
            charges = fields[9]
            radii = fields[10]

            new_line = (f'{startswith} {serial_str} {atomname} {resname} '
                        f'{chid} {alt} {coordinates_x} {coordinates_y} '
                        f'{coordinates_z} {charges} {radii}')
        else:
            new_line = line

        reformat.append(new_line)
    if atom is True:
        prody_pqr = f'{filename[:-4]}_prody.pqr'
        with open(prody_pqr, 'w') as f:
            for new_line in reformat:
                f.write(f"{new_line}\n")
        return prody_pqr
    else: 
        return None


def get_real_sphere(
    out_rna_structure : prody.AtomGroup,
    pocket_structure : prody.AtomGroup,
    analysis : str,
    name : str,
    pqr_out : str,
) -> None:
    """Gets fpocket pocket a-sphere radii form a pqr file and encodes it into
    into the B factor column of a *real_sphere.pdb output file.

    Args:
        pqr_out (string): Path to fpocket *out.pqr file
                           containing a-spheres radii.
        pdb_file (string): Path to fpocket *out.pdb file.
        analysis (str): Path directory containing fpocket outputs for analysis.
        name (str): User specified filename prefix for analysis outputs.
    """
    # PRODY 2.4 can't read PQR files with 4 digit coordinates.
    # reformat_pqr() adds whitespaces between all fields so they can be read. 
    
    if not pocket_structure:
        pqr_out = reformat_pqr(pqr_out)
        if not pqr_out:
            # PQR file contains no ATOM entries (OK).
            writePDB(f'{analysis}/{name}_out_real_sphere.pdb', out_rna_structure)
            return None
        else:
            pocket_structure = parsePQR(pqr_out)
            
        if not pocket_structure:
            # PQR file contains ATOM entries but is not parsed correctly (BAD).
            print(f'\nFAIL unable to reformat PQR file {pqr_out}.')
            writePDB(f'{analysis}/{name}_out_real_sphere.pdb', out_rna_structure)
            return None

        else:
            # PQR file successfully reformated and parsed by Prody 2.4 (OK).
            print(f'Successfully corrected PQR file!\n')

    for residue in out_rna_structure.iterResidues():
        resnum = residue.getResnum()
        resname = residue.getResname()
        if resname == 'STP':
            for atom in residue.iterAtoms():
                coords = atom.getCoords()
                for pqr_atom in pocket_structure.iterAtoms():
                    pqr_resnum = pqr_atom.getResnum()
                    pqr_coords = pqr_atom.getCoords()
                    if resnum == pqr_resnum and (coords == pqr_coords).all():
                        radius = pqr_atom.getRadius()
                        atom.setBeta(radius)
    writePDB(f'{analysis}/{name}_out_real_sphere.pdb', out_rna_structure)


def get_ligand_coords(
    ligand_rna_structure : prody.AtomGroup,
    ligand : str,
    ligandchain : str,
    analysis : str,
    name : str,
) -> tuple[prody.AtomGroup, str]:
    """Gets coordinates and residue name for RNA-binding ligand.

    Args:
        ligand_rna_structure (object): Prody structure of RNA-ligand complex.
        ligand (str): Ligand residue name (usually a 3-letter code).
        ligandchain (str): Chain identifier for desired ligand.
        analysis (str): Path directory contianing fpocket outputs for analysis.
        name (str): Name of input pdb file.

    Returns:
        object: ProDy atomgroup of all atoms in known RNA ligand.
        str: Ligand residue name (usually a 3-letter code).
    """
    if ligand and 2 <= len(ligand) <= 3:
        try:
            ligand_coords = ligand_rna_structure.select(
                f'resname {ligand}').copy()
        except AttributeError:
            print(f'Chain {ligandchain} of {name}.pdb does not contain a '
                  f'ligand named: {ligand}\n'
                  'Please provide a valid ligand chain (--ligandchain) '
                  'and ligand residue name (--ligand).')
            exit()

    elif ligand_rna_structure.select(f'chain {ligandchain} and hetatm '
                                 'and not ion and not water') is None:
        return (None, None)

    else:
        ligand_sele = ligand_rna_structure.select(
            f'chain {ligandchain} and hetatm and not ion and not water').copy()
        hetatm_resn : list = np.unique(ligand_sele.getResnames()).tolist()
    
        if len(hetatm_resn) > 1:
            resnames_qeds : dict = {}
            for resname in hetatm_resn:
                try:
                    response = requests.get(
                        f'https://files.rcsb.org/ligands/download/{resname}_model.sdf')
                    
                    with open(f'{analysis}/{resname}_model.sdf', 'wb') as f:
                        f.write(response.content)
                    mol = Chem.MolFromMolFile(f'{analysis}/{resname}_model.sdf')
                    qed = QED.default(mol)
                    mw : float = Chem.rdMolDescriptors.CalcExactMolWt(mol)
                    pat = Chem.MolFromSmarts("[#6]")
                    num_of_carbon : int =len(mol.GetSubstructMatches(pat))
                    if mw < 100.0 or num_of_carbon <= 2:
                        qed = 0
                        continue
                        
                except:
                    print(f'Error: Not able to calculate QED score for {resname}.\n')
                    qed = 0
                
                resnames_qeds[resname] = qed
                
            # Get ligand with the highest qed score
            hetatm_resn = [max(resnames_qeds, key=resnames_qeds.get)]

        if len(hetatm_resn) == 1 \
                and 2 <= len(hetatm_resn[0]) <= 3:
            ligand_coords = ligand_sele.select(
                f'chain {ligandchain} and resname {hetatm_resn[0]}').copy()
            print(f'Using {hetatm_resn[0]} as ligand for analysis.')
            ligand = hetatm_resn[0]
            return (ligand_coords, ligand)

        elif len(hetatm_resn) == 1 and 2 > len(hetatm_resn[0]) > 3:
            print('No ligand detected.\n')
            return (None, None)

        elif len(hetatm_resn) > 0:
            while True:
                input_resn = input(
                    f'Detected heteroatoms: {hetatm_resn}.\n\n'
                    'Input the target ligand ID (case-sensitive; "none" for no ligand): ')
                print('\n')

                if input_resn in ('n', 'N', 'no', 'No', 'none', 'None'):
                    return (None, None)
                elif input_resn in hetatm_resn:
                    ligand_coords = ligand_sele.select(
                        f'chain {ligandchain} and resname {input_resn}').copy()
                    print(f'Using {input_resn} as ligand for analysis.')
                    ligand = input_resn
                    break
                else:
                    print(f'NameError: {input_resn} is not a not a valid '
                          'ligand ID.\n'
                          'Input "None" to proceed without a ligand.\n')

        else:
            print('No ligand detected.\n')
            return (None, None)


def get_characteristics(
    info_txt : str,
    pdb_code :str,
    name : str,
    m : float,
    M : float,
    i : int,
    D : float,
    A : int,
    p : float,
    state : int,
) -> pd.DataFrame:
    """Creates a dataframe containing characteristics for
        all fpocket generates pockets.

    Args:
        info_txt (string): path to *info.txt file outputed by fpocket.
        pdb_code (string): 4 digit identifier for the PDB structure.
        m (float): Specifies the minimum radius for an a-sphere.
        M (float): Specifies the maximum radius for an a-sphere.
        i (int): Specifies the minimum number of a-spheres per pocket.
        D (float): Specifies the a-sphere clustering distance for pockets.
        A (int): # of electroneg. atoms to define a polar a-sphere.
        p (float): Max. ratio of apolar a-spheres in a pocket.
        state (int): Structural state to analyze.

    Returns:
        DataFrame: Displays the characteristics and properties most relvant
        to scoring each pocket.
    """

    pc_d = {'Parameters': [], 'Name': [], 'PDB': [], 'State': [], 'Pocket': [],
            'Score': [], 'Drug score': [], 'a-sphere': [],
            'SASA': [], 'Volume': [], 'Hydrophobic density': [],
            'Apolar a-sphere proportion': [],
            'Hydrophobicity score': [], 'Polarity score': []}

    with open(info_txt, 'r') as f:

        poc_count = 1

        for row in f:
            if 'Pocket' in row:
                pc_d['Parameters'].append(f'-m {m} -M {M} -i {i} -D {D} '
                                          f'-A {A} -p {p}')
                pc_d['Name'].append(name)
                pc_d['PDB'].append(pdb_code)
                pc_d['State'].append(state)
                pc_d['Pocket'].append(int(row.split(' ')[1].strip()))
                pocket = int(row.split(' ')[1].strip())
            if 'Score :' in row:
                if 'Druggability' not in row:
                    pc_d['Score'].append(float(row.split(':')[1].strip()))
            if 'Druggability Score :' in row:
                pc_d['Drug score'].append(float(row.split(':')[1].strip()))
            if 'Number of Alpha Spheres :' in row:
                pc_d['a-sphere'].append(int(row.split(':')[1].strip()))
            if 'Total SASA :' in row:
                pc_d['SASA'].append(float(row.split(':')[1].strip()))
            if 'Volume :' in row:
                pc_d['Volume'].append(float(row.split(':')[1].strip()))
            if 'Mean local hydrophobic density' in row:
                pc_d['Hydrophobic density'].append(
                    float(row.split(':')[1].strip()))
            if 'Apolar alpha sphere proportion' in row:
                pc_d['Apolar a-sphere proportion'].append(
                    float(row.split(':')[1].strip()))
            if 'Hydrophobicity score' in row:
                pc_d['Hydrophobicity score'].append(
                    float(row.split(':')[1].strip()))
            if 'Polarity score:' in row:
                pc_d['Polarity score'].append(float(row.split(':')[1].strip()))
            if 'Flexibility' in row and poc_count == pocket:
                poc_count += 1

        pc_df = pd.DataFrame.from_dict(pc_d)
        pc_df.insert(loc=4, column='Filter', value='Fail')
        pc_df.insert(loc=4, column='Type', value='Novel')

    return pc_df


def add_basic_characteristics(
    stp_coords : prody.AtomGroup,
    pockets_out : list[str],
    qualityfilter : float,
    pc_df : pd.DataFrame,
    chain :str,
    analysis : str,
    name :str,
    knownnt : list[int],
    ) -> None:
    """Adds characteristics to the pocket characteristics DataFrame that do
        not require a ligand to calculate.
    PocketNT: nucleotides in contact with pocket,
    Geometric Center: X,Y,Z coordinates of pocket center,
    Filter: Pocket quality filter (Pass or Fail)
    
    Args:
        stp_coords (object): ProDy atom group of a-sphere coordinates.
        pockets_out list[str]: Valid paths to pockets/pocket*_atm.pdb files.
        qualityfilter (float): Minimum fpocket score filter for pockets.
        pc_df (DataFrame): Characteristics and properities for each pocket.
        chain (str): Chain identifier for desired RNA chain (default='A').
        analysis (str): path directory contianing fpocket outputs for analysis.
        name (str): Name of input pdb file.
    """
    pocketNT = []
    pocket_npr1 = []
    pocket_npr2 = []

    for i, _ in enumerate(stp_coords.iterResidues()):
        # Export surface obj files for each pocket.
        cmd.load(f'{analysis}/{name}_out_real_sphere.pdb')
        cmd.remove(f'not resn STP or not resi {i+1}')
        cmd.hide('spheres')
        cmd.show('surface')
        cmd.set('surface_quality', 1)
        cmd.save(
            f'{analysis}/pockets/pocket{i+1}_surf.obj',
            f'pocket_{i+1}_surface'
        )
        cmd.reinitialize()

        # Create mesh object for each pocket.
        mesh = trimesh.load(f'{analysis}/pockets/pocket{i+1}_surf.obj')

        # Calc inertia tensor for each pocket.
        inertia = mesh.moment_inertia

        # calculate eigen values and eigen vectors.
        eigvals, _ = np.linalg.eig(inertia)

        # sort eigen values > (I1, I2, I3)
        eig_ord = np.argsort(eigvals)

        # Calculate the normalize PMI ratios for each pocket.
        sorted_eigvals = eigvals[eig_ord]
        I1 = sorted_eigvals[0]
        I2 = sorted_eigvals[1]
        I3 = sorted_eigvals[2]
        npr1 = I1/I3
        npr2 = I2/I3
        pocket_npr1.append(npr1)
        pocket_npr2.append(npr2)

        # Calculate pocketNT (nucleotides near each pocket).
        x = pockets_out[i]
        with open(x, 'r') as f:
            tmp0 = f.read()
            pc_dict = {'Pocket': [(int(x[1]))] for x in findall(
                r'(\w+\s\w+\s*)(\d)(:)(\n)', tmp0)}

            pc_dict.update({x[1].strip(): float(x[2]) for x in findall(
                r'(\w+\s\w+\s*-\s*)(.+):\s*([\d.-]+)(\n)', tmp0)})

        structure = parsePDB(x)

        if ',' in chain:
            chain = chain.replace(',', ' & ')
        selection = structure.select(f'chain {chain}')
        nt = selection.getResnums().tolist()
        pocketNT.append(np.unique(nt).tolist())

    # Add pocketNT and pocket npr data to pc dataframe.
    pc_df['PocketNT'] = pocketNT
    pc_df['Pocket npr1'] = pocket_npr1
    pc_df['Pocket npr2'] = pocket_npr2

    # Add pocket filter (Pass or Fail) to pc dataframe.
    pc_df.loc[pc_df['Score'] > qualityfilter, 'Filter'] = 'Pass'

    # Check if pocketNT matches knownNT
    if isinstance(knownnt, list):
        pc_df['Type'] = pc_df.apply(
            lambda row: 'Known' if sum(nt in knownnt for nt in row['PocketNT']) >= 3 else row['Type'],
            axis=1
        )


def add_ligand_characteristics(
    analysis : str,
    stp_coords : prody.AtomGroup,
    ligand_coords : prody.AtomGroup,
    ligand : str,
    pc_df : pd.DataFrame,
    knownnt : list[int],
) -> None:
    """Adds characteristics to the pocket characteristics DataFrame that
       require the presence of a ligand to calculate.
    Pocket overlap:  Ratio of a-spheres in contact with ligand.
    Ligand overlap:  Ratio of ligand atoms in contact with a-spheres.
    Center criteria: Distance from ligand to the geometric center of pocket.
    Type: Indicates if pocket is known or novel based on:
          pocket overlap, ligand overlap, and center criteria.

    Args:
        analysis (str): path directory contianing fpocket outputs for analysis.
        stp_coords (object): ProDy atom group of a-sphere coordinates.
        ligand_coords (object): ProDy atom group of ligand coordinates.
        ligand (str): Ligand residue name (usually a 3-letter code).
        pc_df (DataFrame): Characteristics and properities for each pocket.
    """

    pocket_overlap = []
    ligand_overlap = []
    center_criteria = []

    # Calculate the normalize PMI ratios for the ligand.
    ligand_npr1, ligand_npr2 = util.calc_npr(ligand_coords)

    #Calculate ligand QED score.
    if not os.path.isfile(f'{analysis}/{ligand}_model.sdf'):
        try:
            response = requests.get(
                f'https://files.rcsb.org/ligands/download/{ligand}_model.sdf')
                
            with open(f'{analysis}/{ligand}_model.sdf', 'wb') as f:
                f.write(response.content)
        except:
            print(f'Error: Not able downlaod ligand model for {ligand}.\n')
            qed = np.nan
    
    try:        
        mol = Chem.MolFromMolFile(f'{analysis}/{ligand}_model.sdf')
        qed = QED.default(mol)
    except:
        print(f'Error: Not able calculate qed score for ligand {ligand}.\n')
        qed = np.nan

    # Iterates through each pocket (residue).
    for _, residue in enumerate(stp_coords.iterResidues()):
        pocket = residue.getCoords()

        # Calculate ratio of a-spheres in each pocket that is near the ligand.
        pocket_overlap_sele = residue.select(
            f'within 3 of ligand', ligand=ligand_coords)
        if pocket_overlap_sele is None:
            pocket_overlap.append(0)
        else:
            # Calculates total number of a-spheres in pocket.
            _, pocket_stp_count = np.unique(
                residue.getResnums(), return_counts=True)

            # Calculates number of overlapped a-spheres in pocket.
            _, pocket_overlap_stp_count = np.unique(
                pocket_overlap_sele.getResnums(), return_counts=True)

            pocket_overlap.append(float(
                pocket_overlap_stp_count / pocket_stp_count))

        # Calculate ratio of a-spheres in each pocket that is near the ligand.
        ligand_overlap_sele = ligand_coords.select(
            f'within 3 of pocket', pocket=residue)
        if ligand_overlap_sele is None:
            ligand_overlap.append(0)
        else:
            # Calculates total number of atoms in ligand.
            ligand_atom_count = ligand_coords.numAtoms()

            # Calculates number of overlapped atoms in ligand.
            ligand_overlap_atom_count = ligand_overlap_sele.numAtoms()

            ligand_overlap.append(
                ligand_overlap_atom_count / ligand_atom_count)
        
        # Calculate minimum distance between pocket center and ligand atoms.
        pocket_center = calcCenter(pocket)
        center_distance_l = calcDistance(pocket_center, ligand_coords)
        center_distance = min(center_distance_l)
        center_criteria.append(center_distance)

    # Add Ligand overlap and Center criteria to pc dataframe.
    pc_df['Pocket overlap'] = pocket_overlap
    pc_df['Ligand overlap'] = ligand_overlap
    pc_df['Center criteria'] = center_criteria

    # Add pocket type (Known or Novel) to pc_df.
    pc_df.loc[
        (pc_df['Ligand overlap'] >= 0.33) &
        (pc_df['Pocket overlap'] >= 0.33) &
        (pc_df['Center criteria'] <= 4), 'Type'
    ] = 'Known'

    # Add ligand NPR and QED score to pc_df.
    pc_df['NPR1'] = np.nan
    pc_df['NPR2'] = np.nan
    pc_df['QED score'] = np.nan
    pc_df.loc[(pc_df['Type'] == 'Known'), 'NPR1'] = ligand_npr1
    pc_df.loc[(pc_df['Type'] == 'Known'), 'NPR2'] = ligand_npr2
    pc_df.loc[(pc_df['Type'] == 'Known'), 'Geometry'] = 'Balanced'
    pc_df.loc[pc_df.eval('NPR1 - NPR2 + 0.5 < 0'), 'Geometry'] = 'Rod-like'
    pc_df.loc[pc_df.eval('- NPR1 - NPR2 + 1.5 < 0'), 'Geometry'] = 'Sphere-like'
    pc_df.loc[pc_df.eval('NPR2 - 0.75 < 0'), 'Geometry'] = 'Disc-like'
    pc_df.loc[(pc_df['Type'] == 'Known'), 'QED score'] = qed



"""Contains all rnavigate convenience plotting functions"""

from fpocketr.rnavigate import plots
from fpocketr.rnavigate import data
from fpocketr.rnavigate.helper_functions import (
    PlottingArgumentParser,
    _parse_plot_kwargs,
    fit_data,
)
from fpocketr.rnavigate.data_loading import get_sequence


__all__ = [
    "plot_options",
    "plot_ss",
]


def plot_options(samples):
    """Prints a list of plotting functions compatible with a sample or list of samples.

    Some plotting functions require specific data classes to be loaded into the
    sample. For plotting multiple samples, data keywords that are not shared, or are
    shared, but are not of the same data class, are considered invalid.

    Parameters
    ----------
    samples : rnavigate.Sample or list of rnavigate.Sample
        samples to check for compatible plotting functions
    """
    if not isinstance(samples, list):
        samples = [samples]
    if len(samples) == 0:
        raise ValueError("No samples provided.")
    # get all data keywords for each sample
    # remove " (default)" from data keyword strings
    # find data keywords that are shared by all samples
    data_keywords = samples[0].print_data_keywords(return_dict=True)
    for key in data_keywords:
        for i, keyword in enumerate(data_keywords[key]):
            keyword = keyword.split(" ")[0]
            data_keywords[key][i] = keyword
    for sample in samples[1:]:
        these_data_keywords = sample.print_data_keywords(return_dict=True)
        for key in data_keywords:
            data_keywords[key] = list(
                set(data_keywords[key]) & set(these_data_keywords[key])
            )
    
    # SecondaryStructures which contain drawing coordinates
    ss_with_coords = []
    for keyword in data_keywords["structures"]:
        if all("X_coordinate" in s.get_data(keyword).data.columns for s in samples):
            ss_with_coords.append(keyword)
    # any sequence data
    sequences = []
    for key in data_keywords:
        for keyword in data_keywords[key]:
            if all(isinstance(s.get_data(keyword), data.Sequence) for s in samples):
                sequences.append(keyword)

    print(
        """
All data keywords are optional inputs for each RNAvigate sample. However,
each plotting function has specific data requirements. This list contains
all plotting functions for which requirements are met by these samples,
and the specific data keywords that satisfy each requirement.

Further below, a list of compatible optional data keywords is provided.
These are likely to be useful for plotting, but are not required.


Compatible plotting functions and keywords for these samples:"""
    )

    if len(ss_with_coords) > 0:
        print(
            f"""
    plot_ss
        structures: {ss_with_coords}"""
        )
    else:
        print(
            """
    plot_ss
    NOT COMPATIBLE
        requires sample(s) which contain structures with coordinates."""
        )

    print("\n\nCompatible optional data for plotting functions\n")
    for key in data_keywords:
        print(f"    {key}: {data_keywords[key]}")


def plot_ss(
    # required
    samples,
    structure,
    # optional data inputs
    profile=None,
    annotations=None,
    interactions=None,
    interactions2=None,
    # optional data display
    labels=None,
    colors=None,
    nt_ticks=None,
    bp_style="dotted",
    # optional plot display
    colorbars=True,
    plot_kwargs=None,
):
    """Generates a multipanel secondary structure drawing with optional
    coloring by per-nucleotide data and display of inter-nucleotide data and/or
    sequence annotations. Each plot may display a unique sample and/or
    inter-nucleotide data filtering scheme.

    Parameters
    ----------
    samples : list of rnavigate Samples
        samples used to retrieve data
    structure : data keyword string or data object
        secondary structure to plot as arcs
        All data are mapped to this sequence before plotting
    profile : data keyword string or data object, defaults to None
        Profile used for coloring if "profile" used in colors dictionary
    annotations : list of data keyword strings or data objects, defaults to []
        Annotations used to highlight regions or sites of interest
    interactions : one of the formats below, defaults to None
        format 1 (data or data keyword)
            Interactions to plot on secondary structure, no filtering
        format 2 (dictionary)
            e.g. {"interactions": format 1}
            additional filtering options can be added to the dictionary
        format 3 (list of format 2 dictionaries)
            This format allows multiple filtering schemes to be applied,
            each will be plotted on a seperate axis
    interactions2 : one of the formats below, defaults to None
        format 1 (data or data keyword)
            Interactions to plot on secondary structure, no filtering
        format 2 (dictionary)
            e.g. {"interactions": format 1}
            additional filtering options can be added to the dictionary
    labels : list of strings, defaults to sample.sample for each sample
        list containing Labels to be used in plot legends
        Defaults to sample.sample for each sample
    colors : dictionary, optional
        a dictionary of element: value pairs that determines how colors
        will be applied to each plot element and if that element is plotted
        only the elements you wish to change need to be included
        value options and what the colors represent:
            None: don"t plot this elelement
            "sequence": nucleotide identity
            "position": position in sequence
            "annotations": sequence annotations
            "profile": per-nucleotide data from profile
                profile argument must be provided
            "structure": base-pairing status
            matplotlib color: all positions plotted in this color
            array of colors: a different color for each position
                must be the same length as structure
        "sequence" may also use "contrast" which automatically chooses
            white or black, which ever contrasts better with "nucleotide"
            color
        Defaults to {"sequence": None,
                        "nucleotides": "sequence",
                        "structure": "grey",
                        "basepairs": "grey"}
    nt_ticks : integer, defaults to None (no labels)
        gap between major tick marks
    bp_style : "dotted", "line", or "conventional", defaults to "dotted"
        "dotted" plots basepairs as a dotted line
        "line" plots basepairs as a solid line
        "conventional" plots basepairs using Leontis-Westhof conventions
            for canonical and wobble pairs ("G-A" plotted as solid dot)
    colorbars : bool, defaults to True
        Whether to plot color scales for all plot elements
    plot_kwargs : dict, defaults to {}
        Keyword-arguments passed to matplotlib.pyplot.subplots

    Returns
    -------
    rnavigate.plots.SS plot
        object containing matplotlib figure and axes with additional plotting and
        file saving methods
    """
    parsed_args = PlottingArgumentParser(
        samples=samples,
        labels=labels,
        structure=structure,
        annotations=annotations,
        profile=profile,
        interactions=interactions,
        interactions2=interactions2,
    )
    plot_kwargs = _parse_plot_kwargs(plot_kwargs, "rnavigate.plots.SS")
    parsed_args.update_rows_cols(plot_kwargs)
    # initialize plot using all structure drawings
    plot = plots.SS(num_samples=parsed_args.num_samples, **plot_kwargs)
    # loop through samples and filters, adding each as a new axis
    for data_dict in parsed_args.data_dicts:
        data_dict = fit_data(data_dict, data_dict["structure"].null_alignment)
        plot.plot_data(**data_dict, colors=colors, nt_ticks=nt_ticks, bp_style=bp_style)
    plot.set_figure_size()
    if colorbars:
        plot.plot_colorbars()
    return plot
